// Revision store implementation.
// Implemented as nhttpd dispatch generators and request handlers.

var file=require('./deps/fs-continuable')
  , E=require('./deps/either'),either=E.either,Right=E.Right,Left=E.Left,isLeft=E.isLeft
  , C=require('./deps/continuable')
  , readFile=file.readFile
  , writeFile=file.writeFile
  , appendFile=file.appendFile
  , streamLines=file.streamLines
  , writeStream=file.writeStream
  , appendStream=file.appendStream
  , bufferedStream=file.bufferedStream
  , staticFile=require('nhttpd/staticFile')
  , handlers=require('nhttpd/handlers')
  , http=require('http')
  , HANDLER=0
  , METHOD=1
  , PATHHEAD=2
  , QUERY=8
  , ChoicePoint=nhttpd.ChoicePoint

function Tree(s,p,ts,m){
  if(typeof ts != 'number')throw new Error('Tree: timestamp must be provided')
  this.locations=[ts,[{parent:p,name:s}]]
  this.manifest=m||p.manifest
  this.timestamp=ts
  this.cn={}
  this.unlinkListeners=[]}

Tree.prototype.isRoot=function(){return this==this.manifest.tree}

Tree.prototype.getOldestPath=function(ts){var l
  if(this.isRoot())return '/'
  l=this.getOldestLocation(ts)
  if(!l.parent)throw new Error("l.parent not found")
  return l.parent.getOldestPath(ts)+l.name}

function _getAt(hist,ts){var i
  if(ts==undefined)return _getCurr(hist)
  if(typeof ts != 'number')throw new Error('ts must be a Number')
  for(i=hist.length;i>0;i-=2) if(hist[i-2]<=ts) return [hist[i-2],hist[i-1],i]
  return [0,null]}

function _getCurr(hist){var l=hist.length
  return l?[hist[l-2],hist[l-1]]:[0,null]}
function _getCurr1(hist){
  return hist[hist.length-1]||undefined}

Tree.prototype.getOldestLocation=function(ts){
  return _getAt(this.locations,ts)[1][0]}

Tree.prototype.getChildNames=function(ts){var ret=[],p,x
  if(ts==undefined)throw new Error('old API')
  for(p in this.cn) if(_getAt(this.cn[p],ts)[1]) ret.push(p)
  return ret}

function _getOrMakeHist(cn,s){
  if(Object.prototype.hasOwnProperty.call(cn,s))return cn[s]
  return cn[s]=[]}

Tree.prototype.getTreeMutating=function(s,ts){var t,hist,x
  if(!ts)throw new Error('old API')
  // for debugging; if call sites are sane this can be commented out.
  if(!isDir(s))throw new Error('Tree.prototype.getTree: '+s+' is a leaf path segment.')
  hist=_getOrMakeHist(this.cn,s)
  if(x=_getCurr1(hist))return x
  t=new Tree(s,this,ts)
  hist.push(ts)
  hist.push(t)
  return t}

Tree.prototype.getTreeExisting=function(s,ts){
  if(!isDir(s))throw new Error('Tree.prototype.getTree: '+s+' is a leaf path segment.')
  if(Object.prototype.hasOwnProperty.call(this.cn,s))return _getAt(this.cn[s],ts)[1]}

Tree.prototype.getLeafExisting=function(s,ts){
  if(isDir(s))throw new Error('Tree.prototype.getTree: '+s+' is not a leaf path segment.')
  if(Object.prototype.hasOwnProperty.call(this.cn,s))return _getAt(this.cn[s],ts)[1]}

Tree.prototype.getLeafMutating=function(s,path,ts){var leaf,hist,x,recordLine
  if(isDir(s))throw new Error('Tree.prototype.getTree: '+s+' is not a leaf path segment.')
  hist=_getOrMakeHist(this.cn,s)
  if(x=_getCurr1(hist))return x
  leaf=new Leaf(s,this,ts)
  leaf.pending={recordLine:'new '+leaf.timestamp+' '+encodeURI(path+s),parent:this,name:s}
  hist.push(leaf.timestamp)
  hist.push(leaf)
  return leaf}

Tree.prototype.injectLeaf=function(s,timestamp){var leaf,hist
  if(isDir(s))throw new Error('Tree.prototype.getTree: '+s+' is not a leaf path segment.')
  leaf=new Leaf(s,this,timestamp)
  hist=_getOrMakeHist(this.cn,s)
  hist.push(timestamp);hist.push(leaf)
  return leaf}

Tree.prototype.injectTree=function(s,timestamp){var t,hist
  if(!isDir(s))throw new Error('Tree.prototype.getTree: '+s+' is a leaf path segment.')
  t=new Tree(s,this,timestamp)
  hist=_getOrMakeHist(this.cn,s)
  hist.push(timestamp);hist.push(t)
  return t}

Tree.prototype.move=function(name,path,destination,ts,replay){var segments,destTree,destHist,curr,sourceHist,source,destName,locations,currLocIndex
  // name is a child of this tree node, destination is a path
  if(typeof ts != 'number')throw new Error('non-numeric timestamp')
  if(typeof name != 'string' || typeof destination != 'string')throw new Error('name and destination must be strings')
  //if(this.isRoot())return [false,"cannot move root"] // XXX this should be allowed, since other such moves-to-subdir-of-self are allowed, but it requires extra work and testing
  // does the thing we are moving exist?
  if(hasprop(this.cn,name))sourceHist=this.cn[name]
  if(!sourceHist)return [false,"move: no such resource "+path+" (in "+this.getOldestPath()+")"]
  source=_getAt(sourceHist,ts)[1]
  if(!source)return [false,"move: no such resource "+path+" at "+ts+" since "+_getAt(sourceHist,ts)[0]]
  // where are we moving it?
  segments=splitPathSegments(destination)
  if(!segments.length)return [false,"move: no destination specified"]
  destName=segments[segments.length-1]
  if(isDir(name)!=isDir(destName))return [false,"move: source ("+name+") and destination ("+destName+") must be both trees or both leaves"]
  if(segments[0]!='/')return [false,"move: destination ("+destination+") must be absolute"]
  sourceHist.push(ts,null)
  // create a function to roll back the change in case we return an error later
  function cancel1(){sourceHist.pop();sourceHist.pop()}
  // find the tree node which is going to be the new location of the resource: if destination = "/foo/bar/baz" we descend into "foo/" and "bar/", starting from the root.
  destTree=_pathWalk(this.manifest.tree,segments.slice(1,-1),ts)
  if(!destTree)return cancel1(),[false,"move: invalid destination "+destination]
  if(_locationIndex(_getAt(source.locations,ts),destTree,destName)>-1)return cancel1(),[false,"move: destination is equal to source"] // XXX a no-op, so maybe just allow it?
  // what is currently in the location to which we are moving it?
  destHist=_getOrMakeHist(destTree.cn,destName)
  curr=_getCurr(destHist)
  if(curr[0]>ts)return [false,"move: change would violate timestamp monotonicity constraint (this should never happen)"]
  // each leaf has a unique timestamp (XXX we should probably test for this where leaves are created) but paths do not, so mov lines use full paths to identify both source and destination
  if(!replay) this.manifest.record('mov '+encodeURI(path)+' '+encodeURI(destination)+' '+ts)
  // there may be unlink listeners on the existing node at the destination, and on the node that is being moved; if so these are fired.
  if(curr[1])_fireUnlink(curr[1])
  _fireUnlink(source)
  locations=_getAt(source.locations,ts)[1]
  locations=_removeFromLocations(locations,this,name).concat([{parent:destTree,name:destName}])
  source.locations.push(ts,locations)
  destHist.push(ts,source)
  _emptyCascade(this,ts)
  return [true]}
// N.B. we now have two names for a tree or leaf, the name in the parent, and the name in the tree or leaf itself.
// There may be multiple references to the same tree or leaf from different (path,time) pairs in tree history, with different names.
// In other words, the name used to reach a resource may not be the only or the current name of the resource.
// Similarly the parent pointer points to the current parent tree, not any previous one.
// Parent pointers could be quantified by time too, but there's (hopefully) no point in doing this if parent pointers are only used in mutating contexts, i.e. only on the current live tree, not any historical tree-segments (gah). [this is now done]

// ([(parent,name)], parent, name) → Int
function _locationIndex(locs,parent,name){var i,l
 for(i=0,l=locs.length;i<l;i++) if(locs[i].parent==parent && locs[i].name==name) return i
 return -1}

function _removeFromLocations(locs,parent,name){var idx,ret
 idx=_locationIndex(locs,parent,name)
 if(idx==-1)throw new Error("failed to find current location index ("+name+" in "+parent.getOldestPath+")")
 ret=locs.slice()
 ret.splice(idx,1)
 return ret}

function hasprop(o,p){return Object.prototype.hasOwnProperty.call(o,p)}

Tree.prototype.link=function(destName,destPath,existPath,ts,replay){var existHist,existResource,existName,existTree,segments,destHist,curr,locations
  if(typeof ts != 'number')throw new Error('non-numeric timestamp')
  if(typeof destName != 'string' || typeof destPath != 'string' || typeof existPath != 'string')throw new Error('link: arguments must be strings')
  // find the existing resource which is being linked to
  segments=splitPathSegments(existPath)
  if(!segments.length)return[false,"link: not a valid link source "+existPath]
  existName=segments[segments.length-1]
  if(isDir(destName)!=isDir(existName))return[false,"link: source and destination must be both trees or both leaves"]
  existTree=_pathWalk(this.manifest.tree,segments.slice(1,-1),ts)
  existHist=existTree.cn[existName]
  if(!existHist)return[false,"link: cannot link to nonexistent resource "+existPath]
  existResource=_getAt(existHist,ts)[1]
  if(!existResource)return[false,"link: no such resource "+existPath+" at "+ts]
  destHist=_getOrMakeHist(this.cn,destName)
  curr=_getAt(destHist,ts)
  if(curr[0]>ts)return[false,"link: monotonicity violation (this should never happen)"]
  if(!replay)this.manifest.record('lnk '+encodeURI(existPath)+' '+encodeURI(destPath)+' '+ts)
  if(curr[1])_fireUnlink(curr[1])
  locations=_getAt(existResource.locations,ts)[1]
  locations=locations.concat([{parent:this,name:destName}])
  existResource.locations.push(ts,locations)
  destHist.push(ts,existResource)
  return [true]}

Tree.prototype.unlink=function(name,path,ts,replay){var nameHist,existing,locations
  if(Object.prototype.hasOwnProperty.call(this.cn,name))nameHist=this.cn[name]
  if(!nameHist)return [true] // rm on non-existent resources silently does nothing (XXX reasonable?)
  if(!(existing=_getAt(nameHist,ts)[1]))return [true]
  if(!replay) this.manifest.record('del '+encodeURI(path)+' '+ts)
  _fireUnlink(existing)
  nameHist.push(ts,null)
  locations=_getAt(existing.locations,ts)[1]
  if(!locations)throw new Error("unlink could not find locations array (existing.locations="+pp(existing.locations)+")")
  locations=_removeFromLocations(locations,this,name)
  existing.locations.push(ts,locations)
  _emptyCascade(this,ts)
  return [true]}

// utility function to fire unlink events on tree or leaf
function _fireUnlink(t_l){var i,l
  for(i=0,l=t_l.unlinkListeners.length;i<l;i++) t_l.unlinkListeners[i]()}

// method shared between Tree and Leaf
Tree.prototype.addUnlinkListener=function(handler){
  this.unlinkListeners.push(handler)}

function _pathWalk(tree,segments,ts){var segment=segments.shift()
  if(tree==undefined)return
  if(segment==undefined)return tree
  if(segment=='/')return _pathWalk(tree.manifest.tree,segments,ts)
  if(segment=='../')throw new Error('_pathWalk called with uncanonicalized path') //return _pathWalk(tree.getParent(ts),segments,ts)
  if(isDir(segment))return _pathWalk(tree.getTreeMutating(segment,ts),segments,ts)
  throw new Error('_pathWalk called with leaf segment')}

// the empty cascade is the effect by which trees are unlinked when they become empty
// it is the dual of the process by which path components are added to the tree as necessary when leaves are created
function _emptyCascade(tree,ts){var locs,i,l
  if(!tree)return
  if(!_emptyTree(tree,ts))return
  locs=_getAt(tree.locations,ts)[1]
  for(i=0,l=locs.length;i<l;i++) _unlink(locs[i].parent,locs[i].name,ts)}

function _unlink(tree,name,ts){var hist,curr
  hist=tree.cn[name]
  curr=_getAt(hist,ts)
  if(curr[0]>ts)throw new Error('this probably should not happen') // XXX
  if(curr[0]==ts)throw new Error('probably should not happen either') // XXX
  if(curr[1]){
    _fireUnlink(curr[1])
    hist.push(ts);hist.push(null)
    _emptyCascade(tree,ts)}}

function _emptyTree(tree,ts){var p
  if(typeof ts != 'number')throw new Error('_emptyTree: typeof ts: '+typeof ts)
  for(p in tree.cn) if(_getAt(tree.cn[p],ts)[1]) return false
  return true}

function isDir(segment){return segment.slice(-1)==='/'}

// splits after every slash except a trailing slash, skipping extra slashes (e.g. "/foo//bar/baz/' → ['/','foo/','bar/','baz/'])
function splitPathSegments(path){var segments,i,l
  segments=path.split(/\/+(?!$)/)
  for(i=0,l=segments.length-1;i<l;i++) segments[i]+='/'
  return segments}

// for better atomicity, needs one timestamp per request, not using +new Date everywhere

// A leaf node contains a timestamp which identifies it uniquely within 
// the manifest.  The timestamp is generated when the leaf is first 
// created, and stored in the manifest record.  It is persistent across 
// the entire life of the resource, including renames.
// A leaf node also contains a list of revisions, which may be 
// serialized on disk as complete contents or as diffs, and which are 
// identified by a time offset from the timestamp of the resource 
// itself.
// Each rev is either a complete file indicated by [0,revtm], or a 
// diff, indicated as [1,revtm].  A consecutive sequence of diffs is 
// stored in a single file, the name of which comes from the revtm 
// of the first diff.
function Leaf(s,p,ts){
  if(typeof ts != 'number')throw new Error('timestamp must be a number')
  this.locations=[ts,[{parent:p,name:s}]]
  this.timestamp=ts
  this.manifest=p.manifest
  this.revs=[]
  this.unlinkListeners=[]}

Leaf.prototype.getOldestPath=Tree.prototype.getOldestPath

Leaf.prototype.getOldestLocation=Tree.prototype.getOldestLocation

Leaf.prototype.isRoot=function(){return false}

Leaf.prototype.addUnlinkListener=Tree.prototype.addUnlinkListener

// Inject a revision, which already exists on the filesystem.  Used when 
// replaying the manifest record.
Leaf.prototype.injectRevision=function(revtm){
  this.revs.push([0,revtm])}

Leaf.prototype.injectDiff=function(revtm){
  this.revs.push([1,revtm])}

Leaf.prototype.commitIfPending=function(){
  if(!this.pending)return
  this.manifest.record(this.pending.recordLine)
  this.pending=undefined}

Leaf.prototype.cancelIfPending=function(){
  if(!this.pending)return
  _unlink(this.pending.parent,this.pending.name,+new Date)
  this.pending=undefined}

// When a diff is stored it is either appended to the existing diff 
// file at the end of the revs array, or a new empty one if one did 
// not exist.
// putDiff (Leaf, String) → Continuable Either Error String
Leaf.prototype.putDiff=function(diff){var t
 t=this
 return function(cont){var fname,revtm,lastrev
  revtm=(new Date)-t.timestamp
  fname=t.currentDiffPath()
  dbg3('putDiff: revtm: '+revtm+' fname: '+fname)
  if(!fname){ // there is no current diff, the last stored revision (if any) must have been complete
    fname=t.fname(revtm)}
  file.appendFile(fname,diff,"binary",0666)
    (E.either(function(e){dbg2('diff write error: '+e);cont(Left(e))}
             ,function(){dbg3('wrote diff '+revtm+' in file '+fname)
                         t.commitIfPending()
                         t.manifest.record('dif '+t.timestamp+' '+revtm)
                         t.revs.push([1,revtm])
                         cont(Right(fname))}))}}

Leaf.prototype.putRevision=function(s){var fname,revtm,t
 t=this
 return function(cont){
  revtm=(new Date)-t.timestamp // timestamp offset of the revision
  fname=t.fname(revtm)
  dbg3(fname)
  // TODO: add a diff branch here, e.g. if the revision is large, the 
  // previous revision is already in memory, and a significantly smaller 
  // diff can be found in < 5ms or so, then write the diff to disk 
  // instead of the full contents of the revision.
  writeFile(fname,s,"binary",0444)
    (E.replaceRight(cont)(function(){
       t.revs.push([0,revtm]) // XXX does this introduce a race condition?
       // revs appear in the order they were finished serializing, not in the order they are received
       t.commitIfPending()
       t.manifest.record('rev '+t.timestamp+' '+revtm)
       cont(E.Right(t.timestamp+revtm))}))}}

// Serialize a new revision from a chunk stream.
// streamRevision :: (Leaf, Stream) → _
Leaf.prototype.streamRevision=function(stream){var leaf=this
  revtm=(new Date)-leaf.timestamp
  fname=leaf.fname(revtm)
  writeStream(fname,'binary',0444)(stream)(E.either(report,succeed))
  // since we have already streamed the result to the user (in revisionPOST), there is not much we can do to indicate whether the write succeeded or not
  // perhaps instead we should return 204 Accepted, and then let the user do another GET if they want to see the retrieved and stored data
  function report(err){
    dump('streamRevision: writing to file '+fname+' failed with: '+err)}
  function succeed(){
    leaf.revs.push([0,revtm])
    leaf.commitIfPending()
    leaf.manifest.record('rev '+leaf.timestamp+' '+revtm)}}

Leaf.prototype.fname=function(revtm){
  return this.manifest.path+this.timestamp+'-'+revtm}

// One and only one of currentRevisionPath and currentDiffPath will 
// return a value for a given leaf at a given time.
Leaf.prototype.currentRevisionPath=function(){var lastrev,l
  l=this.revs.length
  lastrev=this.revs[l-1]
  if(lastrev && lastrev[0]==0)return this.fname(lastrev[1])
  /* otherwise, there is no current revision path (the last thing stored must be a diff) */}

// Get the revision path that was current at the given time
// What we probably want API-wise, in most cases, is a method that returns both a revision path and a set of diffs that must be applied to it
Leaf.prototype.revisionPathAtTime=function(ats){var i,l,rts
  if(typeof ats == 'string')throw new Error('timestamp must be a number')
  rts=ats-this.timestamp // relative timestamp (to resource creation time)
  for(i=this.revs.length;i--;){
    if(this.revs[i][1]<=rts)
      return this.fname(this.revs[i][1])}}

// gets the corresponding index into this.revs for a diff from a given revtm
// The revtm must be exact (not between two existing revtms)
// O(revs.length - returned index)
function _revtm2diffidx(revs,revtm){var i,l
  for(i=revs.length;i--;)
    if(revs[i][0]==1 && revs[i][1]==revtm)
      return i}

// get a diff path, if any, from a given revtm
// i is an optional index into this.revs
// O(distance between i and beginning of corresponding diff file) if i is provided, otherwise O(this.revs.length - index of revtm)
function _revtmDiffPath(leaf,revtm,i){var lastrev,oldest_difftm
  if(i==undefined)i=_revtm2diffidx(leaf.revs,revtm)
  while((lastrev=leaf.revs[i--]) && lastrev[0]==1) oldest_difftm=lastrev[1]
  if(oldest_difftm!=undefined)return leaf.fname(oldest_difftm)}

// TODO: replace these little leaf helper functions with a better designed set

function _revidx2diffstartidx(revs,i){var lastrev,ret
  while((lastrev=revs[i]) && lastrev[0]==1) ret=i--
  return ret}

Leaf.prototype.currentDiffPath=function(){
  return _revtmDiffPath(this,null,this.revs.length-1)}

// Get the last n revision offset timestamps
Leaf.prototype.getLastRevPairs=function(n){var ret=[],l
  for(l=this.revs.length,n=Math.min(l,n);n;n--){
    ret.push(this.revs[l-n])}
  return ret}

// get the entire revision history of a resource, as a sequence of revision timestamps and path moves
Leaf.prototype.getHistory=function(ts){var revs=[],moves=[],i,l,parent,parents,now,lastts,ts,lastpath,path
 now=+new Date
 lastts=0
 for(i=0,l=this.revs.length;i<l;i++){
  assert(this.revs[i][0]===0,'a complete revision, not a diff')
  ts=this.timestamp+this.revs[i][1]
  path=this.getOldestPath(ts)
  if(lastpath!==path){revs.push(path);lastpath=path}
  revs.push(ts-lastts)
  lastts=ts}
 return revs}

// a history array is a list of timestamps (even indices) and items (odd indices)
// it describes the history of some value over time, which changed at history[i] (i even) to the value at history[i+1]
// _hist_lookup returns the subset of the history which covers the time range given
/* unused (and untested) for now
function _hist_lookup(hist,from,to){var i,l,ret=[]
 from=from||0; to=to||Infinity
 i=0;l=hist.length
 while(i<l && hist[i]<=from) i+=2 // skip items older than the range of interest
 if(i>0) ret.push(hist[i-2],hist[i-1]) // the item contemporaneous with the start of the range
 while(i<l && hist[i]<=to){ // all remaining items in the range
  ret.push(hist[i],hist[i+1])
  i+=2}
 return ret}
*/

// If a diff corresponds to the given revtm, return it
// Leaf.prototype.getRawDiff :: (implicit this::Leaf, revtm::Integer) → Continuable Either Error String
Leaf.prototype.getRawDiff=function(revtm){
  return readRawDiff(this,revtm)}

// Read a raw diff from a diff file or cache
// to minimize I/O we read an entire diff file in at a time and cache the diffs in it by timestamp
// currently this cache grows without bound while the manifest is in memory, we probably want to do something about that eventually

// for the case where we are already reading from the file containing this diff (expected to be a common case) we optimize by simply waiting for the existing read to finish.
// this is that rare case where we are actually going to add a callback on an event that is already in progress, i.e. a case where a Promise is more convenient than a pure callback
// If the promise API used here supported stored values, we could do away with the _diffCachePending as well, and handle all these cases the same.
// In fact, using continuations, the cache could be abstracted out of this function, into a more general cache provider:
// makeCache :: (get :: (k → Continuable v)) → Cache k v
// (it's implied here that the cache has state)
// type Cache k v :: { lookup :: k → Continuable v
//                   , apply :: k → (v → a) → Continuable a
//                   }
// However this doesn't really fit here.
// What we want to cache isn't the contents of the file, it's the individual lines, keyed by timestamp.
// Caching the lines individually wouldn't buy us anything because we'd have to break the encapsulation of the cache to get the optimization of not doing duplicate reads.
// So here, perhaps the best abstraction really is the Promise.
// However, we still have to test if there is a pending promise, and otherwise start one.
// So perhaps this itself would be a useful abstraction.
// We don't actually want to cache the file, but if there is a process already reading it, we want to wait until that process is finished.
// Essentially we have an event, which is "the file has been read and the contents are in _diffCache", and we want to queue this event if it is not already going to occur, and then wait for it in either case.

// xxx :: (k → Continuable a) → PendingCache k a
// yyy :: PendingCache k a → k → (a → b) → (PendingCache k a, Continuable b)

// So this is actually just like the cache, except that the values of type a aren't stored, they are just passed to the function a → b.
// This is just like calling the Continuable a with (a → b) directly, except when a pending event for that key is already queued.

// another optimization here would be to only read the end of the file, currently we will read the whole file when a new diff is appended to it and then requested, even if the earlier diffs are already in the cache.
// to do this we need to keep track of byte offsets corresponding to each diff

// readRawDiff :: Leaf, Integer → Continuable Either Error String
function readRawDiff(leaf,revtm){return function(cont){var x,diffidx,diffstartidx,fname,pending
  //dbg3('readRawDiff: '+leaf.getOldestPath()+' '+revtm)
  if(!leaf._diffCache){leaf._diffCache={};leaf._diffCachePending={}}
  if(x=leaf._diffCache[revtm]){dbg3('diff from cache');cont([1,x]);return}
  diffidx=_revtm2diffidx(leaf.revs,revtm)
  diffstartidx=_revidx2diffstartidx(leaf.revs,diffidx)
  fname=leaf.fname(leaf.revs[diffstartidx][1])
  dbg3('looking for diff in '+fname)
  if(!fname){cont([0,'could not find diff file name for '+leaf.getOldestPath(ts)+' revtm:'+revtm]);dbg3('failed to find a file');return}
  function f(){var x
    if(x=leaf._diffCache[revtm])cont([1,x]),dbg3('read diff from file successfully')
    else {dump('corrupt diff file?');cont([0,'could not find revtm '+revtm+' in diff file '+fname])}} // really shouldn't happen, maybe we should throw here also
  if(pending=leaf._diffCachePending[fname]){pending.push(f);dbg3('diff read is pending, cont pushed');return}
  leaf._diffCachePending[fname]=pending=[f] // array of continuations to fire after the read finishes and the cache is populated
  readFile(fname,"binary")(E.either(
    function(e){dump('readRawDiff error '+e);throw e}, // just die if we can't read the file for some reason
    function(data){var i,l
      dbg3('read diff file successfully, calling '+pending.length+' pending callbacks')
      populateLeafDiffCache(leaf,diffstartidx,data)
      for(i=0,l=pending.length;i<l;i++) pending[i]()
      leaf._diffCachePending[fname]=pending=undefined}))}}

function populateLeafDiffCache(leaf,diffstartidx,data){var lines,i,l
  lines=data.split('\n')
  if(!lines.length)return // file was empty
  assert(lines[lines.length-1]=="") // there will be an empty string after the last newline
  lines.pop()
  for(i=0,l=lines.length;i<l;i++){
    if(!leaf.revs[diffstartidx+i]){
      dump(leaf.revs);dump(lines)
      throw new Error('Fatal: more lines in diff file ('+lines.length+') than diff revs in manifest ('+(leaf.revs.length-diffstartidx)+'), one or both may be corrupt')}
    assert(leaf.revs[diffstartidx+i][0]==1) // the rev is a diff
    leaf._diffCache[leaf.revs[diffstartidx+i][1]]=lines[i]}}

// A manifest is initialized by giving an absolute path to a filename 
// which must be readable and writable by the server process.  If the 
// file does not exist, it will be created and the manifest will be 
// empty, otherwise the manifest history record will be read and 
// replayed.

// function Manifest :: (path,ts) → Continuable Either Error Manifest
function Manifest(path,ts){return function(cont){var manifest,recordPath,lineStream
  recordPath=path+"manifest"
  dbg2("Manifest ("+recordPath+")")
  lineStream=streamLines(recordPath,"ascii")
  manifest={}
  manifest.tree=new Tree("/",null,0,manifest)
  manifest.path=path
  // we want to ensure that nothing that happens during the replay appends anything to the manifest record
  // we ensure that here by not creating the record function until the replay is complete
  manifest.injectLeaf=makeInjectLeaf(manifest)
  manifest.injectRevision=makeInjectRev(manifest)
  manifest.injectDiff=makeInjectDiff(manifest)
  manifest.replayMove=makeReplayMove(manifest)
  manifest.replayDel=makeReplayDel(manifest)
  manifest.replayLink=makeReplayLink(manifest)
  manifest.resourceHash={} // used for quick lookups of resources by timestamp on manifest record replay
  // since this is all it is used for, we delete it once the replay is finished
  lineStream(function consumer(m,x){switch(m){
    case 'chunk':dbg2('manifest: '+x);replayManifestRecordLine(manifest,x);lineStream('next');break
    case 'eof':dbg2('manifest eof');finish(manifest);break
    case 'error':cont(Left(x))}})
  function finish(manifest){// after the replay, but before resolving the promise
   manifest.record=makeRecordFunction(manifest)
   manifest.resourceHash=undefined
   cont(Right(manifest))}}}
  
function makeInjectLeaf(manifest){return function(ts,path){var segments
  if(typeof ts != 'number')throw new Error('timestamp argument must be a number')
  segments=path.slice(1) //lose the leading slash
               .split('/')
  go(segments,manifest.tree)
  function go(segments,tree){
    if(segments.length==1)
      manifest.resourceHash[ts]=tree.injectLeaf(segments[0],ts) // store the leaf in the hash for O(1) lookup when adding revisions
    else
      go(segments,tree.getTreeMutating(segments.shift()+'/',ts))}}}

function makeInjectRev(manifest){return function(resource,offset){var leaf
  leaf=manifest.resourceHash[resource]
  if(!leaf){ /* TODO: walk the tree? This probably shouldn't really ever happen */ throw new Error("shouldn't happen") }
  leaf.injectRevision(offset)}}

function makeInjectDiff(manifest){return function(resource,offset){var leaf
  leaf=manifest.resourceHash[resource]
  if(!leaf)throw new Error("no such resource")
  leaf.injectDiff(offset)}}

function makeReplayMove(manifest){return function(from,to,ts){var segments,parent,res
  dbg2('manifest replay move from '+from+' to '+to+' at '+ts)
  segments=splitPathSegments(from)
  assert(segments[0]==='/')
  parent=_pathWalk(manifest.tree,segments.slice(1,-1),ts)
  assert(parent)
  dbg3('manifest replay move parent: '+parent.getOldestPath(ts))
  res=parent.move(segments[segments.length-1],from,to,ts,true)
  if(!res[0])throw new Error('manifest record move line replay failed with: '+res[1])}}

function makeReplayDel(manifest){return function(path,ts){var segments,parent
  segments=splitPathSegments(path)
  assert(segments[0]==='/') // must be absolute
  parent=_pathWalk(manifest.tree,segments.slice(1,-1),ts)
  assert(parent)
  parent.unlink(segments.slice(-1),path,ts,true)}}

function makeReplayLink(manifest){return function(dest,source,ts){var segments,parent,res
  dbg2('manifest replay ln '+dest+' '+source+' at '+ts)
  segments=splitPathSegments(source)
  assert(segments[0]==='/')
  parent=_pathWalk(manifest.tree,segments.slice(1,-1),ts)
  assert(parent)
  dbg3('manifest replay link parent: '+parent.getOldestPath(ts))
  res=parent.link(segments[segments.length-1],source,dest,ts,true)
  if(!res[0])throw new Error('manifest record move line replay failed with: '+res[1])}}

// return a function to record lines in the manifest record.  The lines 
// are queued and then written to disk.
// The fileIO module doesn't have anything as raw as a raw file handle.
// However we do have a writeStream and now a matching appendStream, and 
// the Stream abstraction, which is asynchronous in both directions, is 
// sufficient to allow holding a file descriptor open and streaming to it.
// 
// The incoming line will have a newline appended to it and should not 
// contain any newlines already
function makeRecordFunction(manifest){var recordPath,consumer,queue=""
  recordPath=manifest.path+'manifest'
  appendStream(recordPath,'ascii')
    (bufferedStream(ourStream))
    (function(x){throw new Error('makeRecordFunction: reached the end of an infinite stream '+x)})
  function ourStream(x){
    if(!consumer){assert(typeof x=='function');consumer=x;if(queue)consumer('chunk',queue)}
    if(x=='close')throw new Error('something closed the manifest record stream, bailing out')}
  return function(line){if(!consumer){queue+=line+'\n'}else consumer('chunk',line+'\n');assert(line.indexOf('\n')==-1)}}

// recreate the manifest from the lines of the manifest record
function replayManifestRecordLine(manifest,line){var tokens
  if(!line.length)return // skip blank lines
  assert(line[line.length-1]=='\n')
  line=line.slice(0,-1) // remove newline
  tokens=line.split(' ')
  switch(tokens[0]){
    case 'new':
      manifest.injectLeaf(+tokens[1]            /* timestamp */
                         ,decodeURI(tokens[2])) /* full path */
      break
    case 'rev':
      manifest.injectRevision(+tokens[1]  /* resource timestamp */
                             ,+tokens[2]) /* revision age offset */
      break
    case 'dif':
      manifest.injectDiff(+tokens[1]  /* resource timestamp/id */
                         ,+tokens[2]) /* diff age offset */
      break
    case 'mov':
      manifest.replayMove(decodeURI(tokens[1])   /* from path */
                         ,decodeURI(tokens[2])   /* to path */
                         ,+tokens[3])            /* operation timestamp */
      break
    case 'del':
      manifest.replayDel(decodeURI(tokens[1])   /* deleted resource path */
                        ,+tokens[2])            /* deletion timestamp */
      break
    case 'lnk':
      manifest.replayLink(decodeURI(tokens[1])  /* destination */
                         ,decodeURI(tokens[2])  /* source */
                         ,+tokens[3])           /* timestamp */
      break
    default:
      throw new Error('unknown or corrupted line in manifest record: '+lines[i])}}

// `dispatch` is an exported dispatch generator factory taking two arguments.
// The first is a path to the manifest, the second is a shared state object.
// TODO: add locking on the manifest file (in case a second instance is erroneously launched)
// The manifest path is opened if there is not already a manifest object for that path (or a promise) open in the shared state object.
exports.dispatch=dispatch
// dispatch :: (path,state) → String → Continuable Maybe Pattern
function dispatch(path,state,extend){var gen
 return gen || function _dispatch(pathHead){return function(cont){
   withManifest(path,state,function(manifest){
     gen=makeGenerator({parent:null,name:"/",accessPath:"/"},manifest.tree,undefined,extend)
     gen(pathHead)(cont)})}}}

// We wish to provide access to all resources, and their histories, which have existed at any point in the revision store.
// Internally resources are identified by their timestamp, and externally they are identified by a path, and any given moment in time, which may not be unique over the life of the revision store.
// The most direct way to expose all resources is as a simple list of resource timestamp identifiers, which will identify every resource in the manifest, including those which no longer exist.
// Another way, which may be more user-friendly, is to expose resources which no longer exist or which no longer have the same names in the same way as currently mapped resources.
// That is to say, when a resource is deleted or moved its ghost remains in the original location.
// This applies to both leaf resources and to trees themselves (other than the root which cannot be moved).
// Then in this mode when listing the root we will list the the trees and leaves which have at any time been present at the root.
// Those which are no longer present will include some notation to that effect in the listing (perhaps a timestamp at which they were deleted or moved, and in the latter case the new location) (A UI could use color to provide hints as to the current disposition and age of such ghosts, e.g. red for those which were deleted, some other color for those which were moved, with the resource fading to become less visible as it ages.)
// A time range could be given, and the listing could then be limited to resources which existed in the given range.
// This is similar to timeTravelGET, but travelling into a time range rather than a particular instant.
// The /asof/ URI space can be used to cover this purpose, then, by giving timestamp ranges.

// Instead of time slices, we wish to expose the entire revstore history in one tree under the -history/ branch.
// Unlike the -asof branch, or the planned time range branch, the history branch is not immutable, since it covers the entire history up to the present.
// For each resource we return a list of revision timestamps, which can be used to retrieve the revision from the -asof branch.
// For directories, we need to return a list of the current resources, those that were deleted or moved away, but also any resource that used to exist at that path, even if the tree object it was in was subsequently moved away.
// Actually, this can all be layered on top so all we really need is:
// Each leaf resource can be queried for its complete revision history.
// Every directory in the path can also be queried for a move of the relevant resource.
// So for the file /foo/bar/baz, we start at / and look at the current value of foo/, and how old it is, and where it came from.
// It was either created, or moved or cloned from somewhere else, at a particular time.
// Then we look in foo/ at bar/ and get the same information.
// If bar/ is more recent than foo/, then we throw away the foo/ data, as a move that happened to foo/ when bar/ was not part of it isn't relevant to the history of the foo/bar/baz resource.
// Looking up baz in bar/ we then do the same comparison, and are left with the current /foo/bar/baz resource and the timestamp of its most recent change, either creation or a move from a different location.
// For a history feature we don't even need this much.
// If the history feature doesn't use paths at all, it'll be easier.
// So: each leaf resource has a unique id.
// Each tree is identified by the path and timestamp at which it was created.
// So for leaf resources we can have /code-history/{timestamp}, and for directories /code-history/{timestamp}{path}
// then there's the problem of getting the resource id from the current URI anyway...
// So.
// We do a single GET and we get a list of every move or branch, and every revision timestamp.
// This is sufficient to look up everything in -asof back to the creation of the resource.
// To look up deleted resources, 
// instead of -history, we could just do ?history

// Moving is supported using the MOVE method, and always succeeds, replacing any existing resource at the moved-to location.
// As in WebDAV, a Destination header is required with the MOVE method.
// Only relative URI references are supported values for this header.

// There will be no copy, since to copy the entire history of a resource (unless to a different host or revision store, of course) would be wasteful and useless, but there will be a branch operation.
// However there may not need to be much support for this in the revision store itself, if the next higher layer supports a 'branch from' operation, then this can be implemented simply by creating a new resource with the first operation on that resource being a 'branch from' operation referring to an existing revision.
// This higher layer can then provide UI (which will necessarily be media-type specific anyway) to deal with watching for and perhaps optionally merging new changes to the resource of which this is a branch (in either direction).

// Deletion is supported by DELETE.
// Subsequent requests will see 404.

// Read-only access to the revision store as it existed at any time in the past (since the creation of the manifest)
// The first path component is the timestamp
exports.timeTravelGET=timeTravelGET
// :: (path, state) → (String) → Continuable Maybe Pattern
function timeTravelGET(path,state){return function _timeTravelGET(pathHead){return function(cont){var m,ts
   if(!(m=/^(\d+)\/$/.exec(pathHead)))return cont()
   ts=+m[1]
   if(ts>+new Date)return cont() // serving data from the future not yet implemented
   withManifest(path,state,function(manifest){
     cont({test:pathHead
          ,then:[PATHHEAD,[{generator:makeGenerator({parent:null,name:"/",accessPath:"/"},manifest.tree,ts,false,true)}]]})})}}}

// withManifest takes the path of a manifest, a state object, and a function
// it looks up the manifest with that path in the state object, creates it if missing, and once it is created, calls the function
// it returns either the return value of the function immediately (if the manifest already existed), or a promise otherwise
function withManifest(path,state,f){var cached,manifestPromise,ts
 if(!state.manifests)state.manifests={}
 if(cached=state.manifests[path]){
  if(cached.pending==true) cached.callbacks.push(f)
  else f(state.manifests[path])} // this is the only case in which we call f immediately
 else{ // nothing is cached
  if(!isDir(path))throw new Error("revisionDispatch: path must end with a slash") // TODO: resolve relative paths to absolute
  ts=+new Date
  manifestPromise={pending:true,callbacks:[f]}
  Manifest(path)(either(lose,win))
  function win(m){
   dump('created manifest in '+(new Date-ts)+' ms')
   state.manifests[path]=m
   manifestPromise.callbacks.forEach(function(f){f(m)})}
  function lose(e){
   dump('manifest creation error '+e)}
  state.manifests[path]=manifestPromise}} // cache the promise, so that another invocation will not also try to open the same manifest

// makeGenerator :: ( location::(parent,name,accessPath), tree::Tree, ts::Integer, extend::Boolean, safe::Boolean ) → String → Continuable Maybe Pattern
// makes a dispatch generator for tree
// if ts is set, descends into the tree as it existed in the past at time ts, otherwise uses the current time
// if 'extend' is true, ts should not be set, and the generator should accept requests for non-existent resources by creating them
// if neither 'extend' nor 'safe' is true, the generator will first serve an existing resource, but will then branch on the method, extending the tree if the method is PUT or POST
// if 'safe' is true, the generator will not extend the tree and will only accept GET requests
function makeGenerator(loc,tree,ts,extend,safe){return function generator(pathHead){return function(cont){var subtree,leaf,pattern,subpath,subloc
   subpath=loc.accessPath+pathHead
   subloc={parent:tree,name:pathHead,accessPath:subpath}
   if(isDir(pathHead)){
    if(subtree=tree.getTreeExisting(pathHead,ts)){
     pattern={test:pathHead
             ,then:pathBranch(subloc,subtree,ts,safe)}
     subtree.addUnlinkListener(function(){pattern.destroy()})
     cont(pattern)}
    else{
     if(!extend)return cont() // decline
     ts=+new Date
     subtree=tree.getTreeMutating(pathHead,ts)
     pattern={test:pathHead,transient:true
             ,then:[PATHHEAD,[{generator:makeGenerator(subloc,subtree,ts,extend)}]]}
     cont(pattern)}}
   else if(pathHead==''){
    pattern={test:pathHead
            ,then:listingBranch(loc,tree,ts,safe)}
    tree.addUnlinkListener(function(){pattern.destroy()})
    cont(pattern)}
   else{
    if(leaf=tree.getLeafExisting(pathHead,ts)){
     pattern={test:pathHead
             ,then:leafBranch(subloc,leaf,ts,safe)}
     leaf.addUnlinkListener(function(){dbg1('got unlink event, destroying pattern for '+leaf.getOldestPath(ts));pattern.destroy()})
     cont(pattern)}
    else{
     if(!extend)return cont() // decline
     ts=+new Date
     leaf=tree.getLeafMutating(pathHead,loc.accessPath,ts)
     pattern={test:pathHead,transient:true
             ,then:leafBranch(subloc,leaf,ts)}
     cont(pattern)}}}}}

function pathBranch(loc,tree,ts,safe){
 if(safe) return [PATHHEAD,[{generator:makeGenerator(loc,tree,ts,false,true)}]]
 else     return [PATHHEAD,[{generator:makeGenerator(loc,tree,ts)}
                           ,{catchall:true
                            ,then:[METHOD,[{regex:/^(?:PUT|POST)$/
                                           ,then:[PATHHEAD,[{generator:makeGenerator(loc,tree,ts,true)}]]}
                                          ,{catchall:true
                                           ,then:[HANDLER,handlers.handle404]}]]}]]}

function listingBranch(loc,tree,ts,safe){
 if(safe) return [METHOD,[{test:'GET'
                          ,then:treeListing(tree,ts)}]]
 else     return [METHOD,[{test:'GET'
                          ,then:treeListing(tree,ts)}
                         ,{test:'DELETE'
                          ,then:[HANDLER,treeUnlinker(loc)]}
                         ,{test:'MOVE'
                          ,then:[HANDLER,treeMover(loc)]}
                         ,{test:'POST'
                          ,then:[HANDLER,treePOST(loc)]}]]}

function leafBranch(loc,leaf,ts,safe){
 if(safe) return [METHOD,[{test:'GET'
                          ,then:revisionGET(leaf,loc,ts)}]]
 else     return [METHOD,[{test:'GET'
                          ,then:revisionGET(leaf,loc)}
                         ,{test:'PUT'
                          ,then:[HANDLER,revisionPUT(leaf)]}
                         ,{test:'POST'
                          ,then:[HANDLER,revisionPOST(loc,leaf)]}
                         ,{test:'DELETE'
                          ,then:[HANDLER,treeUnlinker(loc)]}
                         ,{test:'MOVE'
                          ,then:[HANDLER,treeMover(loc)]}]]}

function treeListing(tree,ts){
 return [QUERY,[{test:''
                ,then:[HANDLER,treeListing_HTML(tree,ts)]}
               ,{test:'text'
                ,then:[HANDLER,treeListing_plain(tree,ts)]}]]} // TODO: JSON

function treeListing_plain(tree,ts){return function(req,resp){var body,inner_ts
  inner_ts=ts||+new Date
  body=tree.getChildNames(inner_ts).join('\n')
  resp.writeHead(200,{"Content-Type":"text/plain"})
  resp.write(body)
  resp.end()}}

function treeListing_HTML(tree,ts){return function(req,resp){var body,inner_ts
  inner_ts=ts||+new Date
  body='<ul>'
      +(tree.isRoot()?'':'<li><a href=..>..</a>')
      +tree.getChildNames(inner_ts).map(function(name){return treeToLI(name)}).join('\n')
      +'</ul>'
  resp.writeHead(200,{"Content-Type":"text/html"})
  resp.write(body)
  resp.end()}}

function treeToLI(name){var query // not actually a tree, but it is in spirit
  // XXX actually need to escape quotes here, not just & and <
  //query=name[name.length-1]=='/'?'?html':''
  return '<li><a href="'+htmlEntify(name)+'">'+htmlEntify(name)+'</a>'}

function treeMover(loc){return function(req,resp){var res
  res=loc.parent.move(loc.name,loc.accessPath,req.headers['destination'],+new Date)
  if(res[0]){resp.writeHead(200,{'Content-Type':'text/plain'});resp.write('Moved.')}
  else{resp.writeHead(403,{'Content-Type':'text/plain'});resp.write(res[1])}
  resp.end()}}

function treeUnlinker(loc){return function(req,resp){var res
  res=loc.parent.unlink(loc.name,loc.accessPath,+new Date)
  if(res[0]){resp.writeHead(200,{'Content-Type':'text/plain'});resp.write('Removed.');resp.end()}
  else{resp.writeHead(403,{'Content-Type':'text/plain'});resp.write(res[1]);resp.end()}}}

function treePOST(loc){return function(req,resp){var xlinkwith,re_rooted,res
  xlinkwith=req.headers['x-link-with']
  dbg2('X-Link-With: '+xlinkwith)
  res=loc.parent.link(loc.name,loc.accessPath,xlinkwith,+new Date)
  dbg2('result: ');dbg2(res)
  if(res[0]){resp.writeHead(200,{'Content-Type':'text/plain'});resp.write('Linked.');resp.end()}
  else{resp.writeHead(403,{'Content-Type':'text/plain'});resp.write(res[1]);resp.end()}}}

function revisionGET(leaf,loc,ts){
 return [QUERY,[{test:"history"
                ,then:[HANDLER,revisionHistory(leaf,ts)]}
               ,{test:""
                ,then:[HANDLER,_revisionGET]}]]
 function _revisionGET(req,resp){var fname,name,ext
  if(ts==undefined)fname=leaf.currentRevisionPath()
  else fname=leaf.revisionPathAtTime(ts)
  ext=loc.name.slice(loc.name.lastIndexOf('.')+1)
  if(!fname)sendEmpty(req,resp)
  else staticFile._staticFileGET(fname,{ext:ext,mime_default:'text/plain'})(req,resp)}}

function revisionHistory(leaf,ts){return function _revisionHistory(req,resp){var history,body
 //ts=ts||+new Date
 history=leaf.getHistory(ts)
 dbg2({'history':history})
 body=JSON.stringify(history)
 resp.writeHead(200,{"Content-Type":"application/json"
                    ,"Content-Length":body.length})
 resp.write(body)
 resp.end()}}

function sendEmpty(req,resp){
  resp.writeHead(200,{"Content-Length":0})
  resp.end()}

function revisionPUT(leaf){return function _revisionPUT(req,resp){
  getRawBody(req,resp,function(body){
    leaf.putRevision(body)(either(lose,win))
    function win(ts){var body
      body=''+ts
      resp.writeHead(200,{'Content-Type':'text/plain'
                         ,'Content-Length':body.length})
      resp.write(body)
      resp.end()}
    function lose(e){
      handlers.error(500,"Server error","revisionPUT failed "+e)(req,resp)}})}}

// Here we handle POST requests to a resource that is considered as a 'log'.
// The most common type of operation on a log is to append a line to it.
// Log lines are imagined as being appended atomically and instantaneously; the appended line is expected to be given entire in one request.
// XXX needs to be replaced or folded in to the rest of the revstore dispatch tree and update handling.
exports.logPOST=logPOST1
// logPost :: (String, String, Object) → Continuable Either Error Handler
function logPOST1(resource,path,state){return function(cont){
  withManifest(path,state,function(manifest){
    cont(Right(logPOST2(manifest.tree.getLeafMutating(resource,path,+new Date))))})}}

function logPOST2(leaf){return function _logPOST2(req,resp){
  getRawBody(req,resp,function(body){
    // TODO: add some test that this is actually a diff (in whatever format we require)
    // TODO: test that the Content-Type is correct (which of course belongs not here at all but in the dispatch tree)
    body=body.replace(/\n/g,'\\n')
    if(!body.length){
      resp.writeHead(400,{})
      resp.end()
      return}
    body=body+'\n'
    leaf.putDiff(body)(either(lose,win))
    function win(){
      resp.writeHead(204,{})
      resp.end()}
    function lose(){
      handlers.error(500,"Server Error","logPOST2: error writing diff")(req,resp)}})}}

// The tail of a log can be retrieved as an HTML page
// logTail :: (String, String, String, Object, Integer) → Continuable Either Error Handler
exports.logTail=logTail
function logTail(resource,asof_relURI,path,state,limit){return function(cont){
  withManifest(path,state,function(manifest){
    cont(Right(logTail2(manifest.tree.getLeafExisting(resource),asof_relURI,limit)))})}}

// logTail2 :: (Leaf, limit::Integer) → RequestHandler
// sends the last `limit` diffs from the requested log to the client as an HTML page consisting of a single <ul>
function logTail2(leaf,asof_relURI,limit){
  if(limit==undefined)limit=10
  if(limit==-1)limit=Infinity
  return function _logTail2(req,resp){var revpairs,ts,continuables
    revpairs=leaf.getLastRevPairs(limit)
    continuables=revpairs.map(function(revpair){
      if(revpair[0] != 1) throw new Error("logTail2: request for non-diff revision not yet implemented") // XXX
      return leaf.getRawDiff(revpair[1])})
    dbg3('logTail2: reading '+continuables.length+' raw diffs')
    C.parallel(continuables)(function(diffs /*[Either Error String]*/){var i,l,diff,lis,ats
      dbg3('logTail2: got '+diffs.length+' results')
      lis=[]
      for(i=0,l=diffs.length;i<l;i++){diff=diffs[i]
        if(isLeft(diff)) throw new Error('TODO: something reasonable here') // XXX
        ats=leaf.timestamp+revpairs[i][1]
        lis[i]=logDiffToLI(diff[1],ats,asof_relURI)}
      serveLogTail(lis,req,resp)})}}

function logDiffToLI(diff,ts,asof_relURI){
  return '<li><a href='+asof_relURI+ts+'/>'+htmlEntify(diff.slice(2))+'</a>'}

function serveLogTail(lis,req,resp){
  resp.writeHead(200,{"Content-Type":"text/html"})
  resp.write("<ul>\n"+lis.join('\n')+'</ul>\n','binary')
  resp.end()}

// We need a way to GET a request from a remote server and put the response in the revision store.
// One way to do this is to do the GET on the client, and then just use an ordinary PUT to store it.
// However this fails if the client is a browser, because client JavaScript cannot do an arbitrary GET request because of the same-origin policy.
// This can be worked around with a proxy, which means that getting something from a remote HTTP server into the revision store would require a request from the browser to the local server, which would do a request to the remote server, return the result to the client, and the client would then do a PUT, so the same response body is transmitted three times, and there are a total of at least three HTTP requests.
// Furthermore if the response is very large then the process may not even succeed.
// So, we add the direct capability to GET a remote resource and store the returned representation in the revision store in a single step.

// When nhttpd supports modules, this can be split out into its own module.

// This is initiated by doing a POST request, with an X-Request-From: header which contains a URI.
// If the headers are considered interesting, then they can be stored as well by giving an extra path with X-Store-Headers-At: uri-path
// Other headers could be used to specify things like what happens with a 404, etc.

// a POST can do pretty much anything, there may be several submodules that allow doing different things (like appending a diff, etc)
// The test for X-Get-From or whatever else distinguishes these should of course be in the dispatch tree.
function revisionPOST(loc,leaf){return function _revisionPOST(req,resp){var fromURI,xlinkwith
  fromURI=req.headers['x-request-from']
  xlinkwith=req.headers['x-link-with']
  if(fromURI)return revisionPOST_x_request_from(leaf,fromURI,req,resp)
  if(xlinkwith)return revisionPOST_x_link_with(loc,leaf,xlinkwith,req,resp)
  leaf.cancelIfPending()
  handlers.error(400,'Bad Request','Ill-formed or forbidden POST request')(req,resp)}}

function revisionPOST_x_request_from(leaf,fromURI,req,resp){var uri,consumer
 leaf.streamRevision(bufferedStream(clientChunkStream))
 // Leaf.prototype.streamRevision :: (Leaf, Stream) → _
 function clientChunkStream(x){var request
   if(!consumer){
     if(typeof x != 'function')throw new Error('First argument of first call to stream must be a consumer')
     consumer=x
     request=http.createClient(uri.port||80,uri.domain).request('GET',uri.path,{'Host':uri.domain,'Accept':'*/*'})
     request.addListener('response',function(response){
       resp.writeHead(response.statusCode,{'Content-Type':response.headers['content-type']})
       response.setEncoding('binary')
       response.addListener('data',function(chunk){
         consumer('chunk',chunk)
         resp.write(chunk,'binary')})
       .addListener('end',function(){
         consumer('eof')
         resp.end()})})
     request.end()}}}

// this is somewhat cleaner than the previous function, but all the nesting is still ugly and it was hard to write
// a utility function to easily create a buffered stream in this pattern would help
// e.g: simpleBufferedStream :: (Consumer StreamMsg → _) → Stream
// e.g: simpleBufferedStream(function(consumer){while(sleep(10)) consumer('chunk','hi there!')})
// simpleBufferedStream then takes care of buffering and handles the next and close messages from the final consumer
// One benefit here is that we don't need the ready event, since we pass in a stream (which adds laziness)
// A higher-level abstraction that would help even more is a way to treat the HTTP client request as a stream

function revisionPOST_x_link_with(loc,leaf,xlinkwith,req,resp){
 res=loc.parent.link(loc.name,loc.accessPath,xlinkwith,+new Date)
 dbg2('result: ');dbg2(res)
 if(res[0]){resp.writeHead(200,{'Content-Type':'text/plain'});resp.write('Linked.');resp.end()}
 else{leaf.cancelIfPending();resp.writeHead(403,{'Content-Type':'text/plain'});resp.write(res[1]);resp.end()}}

// TODO: use the URI module that somebody already wrote for node
var parseURI_re=/(https?):\/\/([^:/?#]+)(?::(\d+))?(\/[^?#]*)(?:\?([^#]+))?(?:#(.*))?/

function parseURI(uri){var m
  m=parseURI_re.exec(uri)
  if(!m)return
  return {protocol:m[1]
         ,domain:m[2]
         ,port:m[3]
         ,path:m[4]
         ,query:m[5]
         ,fragment:m[6]}}

// XXX this is one get.*Body function too many
function getRawBody(req,resp,f){var body=[]
  if(req.headers['expect'] && req.headers['expect'].indexOf('100-continue')!==-1){
    req.connection.setNoDelay()
    req.connection.write('HTTP/1.1 100 Continue\r\n\r\n')}
  req.setEncoding("binary")
  req.addListener("data",function(chunk){
    body.push(chunk)})
  req.addListener("end",function(){
    body=body.join('')
    f(body)})}

// XXX from ed/generator.js
function htmlEntify(s){
  return s.replace(/[<&]/g,function(x){return{'<':'&lt;'
                                             ,'&':'&amp;'}[x]})}
