// Either is a type (borrowed, loosely, from Haskell)

// If A and B are types, then a value of type Either A B is either a value of type A (called a left value) or a value of type B (a right), along with a flag of some kind that allows you to determine whether you have a left or a right.
// Note that A and B could be the same type or overlapping types, so without the flag it would not necessarily be possible to distinguish left or right values.
// Eithers are commonly used in reporting success or failure; traditionally Left represents a failure or error condition, while Right represents success.

// constructors

exports.Left=Left
exports.Right=Right

// Left :: a → Either a b
function Left(x){return [0,x]}
// Right :: b → Either a b
function Right(x){return [1,x]}

// tests
exports.isLeft=isLeft;exports.isRight=isRight
function isLeft(either){return either[0]===0}
function isRight(either){return either[0]===1}

function content(either){return either[1]}

// higher-order functions on Either

exports.either=either
exports.replaceLeft=replaceLeft
exports.replaceRight=replaceRight
exports.__applyLeft=applyLeft // exported with underscores since these are reused names with completely different semantics
exports.__applyRight=applyRight

// either :: (a → x) → (b → x) → (Either a b → x)
function either(l,r){return function(either){return (either[0]?r:l).apply(null,either.slice(1))}}

// replaceLeft :: (Either a b → x) → (c → x) → (Either c b → x)
function replaceLeft(f_ab){return function(f_c){return function(either){
  if(either[0])return f_ab(either)
  return f_c(either[1])}}}

// replaceRight :: (Either a b → x) → (c → x) → (Either a c → x)
// replace the "right half" of a function taking an either type
function replaceRight(f_ab){return function(f_c){return function(either){
  if(either[0])return f_c(either[1])
  return f_ab(either)}}}

// applyLeft :: (a → c) → Either a b → Either c b
function applyLeft(f){return function(e){
  return e[0]?e:[0,f(e[1])]}}

// applyRight :: (b → c) → Either a b → Either a c
function applyRight(f){return function(e){
  return e[0]?[1,f(e[1])]:e}}